import numpy as np

from pypop7.optimizers.de.de import DE


class JADE(DE):
    """Adaptive Differential Evolution (JADE).

    Reference
    ---------
    Zhang, J., and Sanderson, A. C. 2009.
    JADE: Adaptive differential evolution with optional external archive.
    IEEE Transactions on Evolutionary Computation, 13(5), 945–958.
    https://doi.org/10.1109/TEVC.2009.2014613
    """
    def __init__(self, problem, options):
        DE.__init__(self, problem, options)
        self.mu = options.get('mu', 0.5)  # The initial mean of normal distribution
        self.median = options.get('median', 0.5)  # The initial location parameter of the Cauchy distribution
        self.s = options.get('s', 0.1)  # The scale parameter of the Cauchy distribution
        self.p = options.get('p', 0.05)  # p determines the greediness of the mutation strategy
        self.c = options.get('c', 0.1)  # 1/c is the life span of a successful crossover probability or mutation factor
        self.archive = options.get('archive', True)
        self.boundary_constraint = options.get('boundary_constraint', False)

    def initialize(self, args=None):
        x = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary,
                                            size=(self.n_individuals, self.ndim_problem))
        y = np.empty((self.n_individuals,))
        for i in range(self.n_individuals):
            if self._check_terminations():
                break
            y[i] = self._evaluate_fitness(x[i], args)
        a = np.empty((0, self.ndim_problem))  # the set of archived inferior solutions
        self._n_generations += 1
        return x, y, a

    def mutate(self, x=None, y=None, a=None):
        x_mu = np.empty((self.n_individuals,  self.ndim_problem))
        mu_f = np.empty((self.n_individuals,))  # mutation factors
        order = np.argsort(y)[:int(np.ceil(self.p * self.n_individuals))]  # Choose the 100p% best individuals
        x_p = x[self.rng_optimization.choice(order, (self.n_individuals,))]
        x_un = np.copy(x)
        if self.archive:
            x_un = np.vstack((x_un, a))  # The union of the current population and the archive
        for i in range(self.n_individuals):
            mu_f[i] = self.s * self.rng_optimization.standard_cauchy() + self.median
            while mu_f[i] <= 0:
                mu_f[i] = self.s * self.rng_optimization.standard_cauchy() + self.median
            if mu_f[i] > 1:
                mu_f[i] = 1
            r1 = self.rng_optimization.choice(np.setdiff1d(np.arange(self.n_individuals), i))
            r2 = self.rng_optimization.choice(np.setdiff1d(np.arange(len(x_un)), np.union1d(i, r1)))
            x_mu[i] = x[i] + mu_f[i] * (x_p[i] - x[i]) + mu_f[i] * (x[r1] - x_un[r2])

            if self.boundary_constraint:
                idx = np.array(x_mu[i] < self.lower_boundary)
                if idx.any():
                    x_mu[i][idx] = (self.lower_boundary[idx] + x[i][idx]) / 2
                idx = np.array(x_mu[i] > self.upper_boundary)
                if idx.any():
                    x_mu[i][idx] = (self.upper_boundary[idx] + x[i][idx]) / 2
        return x_mu, mu_f

    def crossover(self, x=None, x_mu=None):
        cr_p = self.rng_optimization.normal(self.mu, 0.1, (self.n_individuals,))  # crossover probabilities
        x_cr = np.copy(x)
        for i in range(self.n_individuals):
            for j in range(self.ndim_problem):
                if (j == self.rng_optimization.integers(self.ndim_problem)) or \
                        (self.rng_optimization.random() < cr_p[i]):
                    x_cr[i, j] = x_mu[i, j]
        return x_cr, cr_p

    def select(self, args=None, x=None, y=None, a=None, x_cr=None, mu_f=None, cr_p=None):
        s_f = []  # the set of all successful mutation factors
        s_p = []  # the set of all successful crossover probabilities
        for i in range(self.n_individuals):
            if self._check_terminations():
                break
            y_temp = self._evaluate_fitness(x_cr[i], args)
            if y_temp < y[i]:
                x[i] = x_cr[i]
                y[i] = y_temp
                if self.archive:
                    a = np.vstack((a, x_cr[i]))
                s_f.append(mu_f[i])
                s_p.append(cr_p[i])
        return x, y, a, s_f, s_p

    def iterate(self, args=None, x=None, y=None, a=None):
        x_mu, mu_f = self.mutate(x, y, a)
        x_cr, cr_p = self.crossover(x, x_mu)
        x, y, a, s_f, s_p = self.select(args, x, y, a, x_cr, mu_f, cr_p)

        if self.archive:
            if len(a) > self.n_individuals:  # randomly remove solutions from a so that |a| <= self.n_individuals
                a = np.delete(a, self.rng_optimization.choice(len(a), (len(a) - self.n_individuals,), False), 0)
        self.mu = (1 - self.c) * self.mu + self.c * np.mean(s_p)  # Update the mean of Normal distribution
        if np.sum(s_f) != 0:  # Update the location of Cauchy distribution
            self.median = (1 - self.c) * self.median + self.c * np.sum(np.power(s_f, 2)) / np.sum(s_f)
        return x, y, a

    def optimize(self, fitness_function=None, args=None):
        fitness = DE.optimize(self, fitness_function)
        x, y, a = self.initialize(args)
        fitness.extend(y)
        while True:
            x, y, a = self.iterate(args, x, y, a)
            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            self._n_generations += 1
            self._print_verbose_info(y)
        results = self._collect_results(fitness)
        return results



import time
import numpy as np

from optimizers.nes.nes import NES


class SNES(NES):
    """Separable Natural Evolution Strategy (SNES).

    Reference
    ---------
    Schaul, T., Glasmachers, T. and Schmidhuber, J., 2011, July.
    High dimensions and heavy tails for natural evolution strategies.
    In Proceedings of Annual Conference on Genetic and Evolutionary Computation (pp. 845-852).
    https://dl.acm.org/doi/abs/10.1145/2001576.2001692

    http://schaul.site44.com/code/snes.py    (see the official Python version)
    """
    def __init__(self, problem, options):
        NES.__init__(self, problem, options)
        if self.eta_mu is None:  # learning rate of mean of Gaussian search distribution (η_μ)
            self.eta_mu = 1
        if self.eta_sigma is None:  # learning rate of std of Gaussian search distribution (η_σ)
            self.eta_sigma = (3 + np.log(self.ndim_problem)) / (5 * np.sqrt(self.ndim_problem))
        self._upsilon = self._fitness_shaping()

    def initialize(self):
        s = np.empty((self.n_individuals, self.ndim_problem))  # samples (population / candidate solutions)
        mu = self._initialize_mu()  # mean of Gaussian search distribution
        sigma = np.ones((self.ndim_problem,))  # individual step-sizes (std of Gaussian search distribution)
        y = np.empty((self.n_individuals,))  # fitness (no evaluation)
        return s, mu, sigma, y

    def iterate(self, s=None, mu=None, sigma=None, y=None, args=None):
        for k in range(self.n_individuals):  # sample population (candidate solutions)
            if self._check_terminations():
                return s, y
            # draw sample (individual)
            s[k] = self.rng_optimization.standard_normal((self.ndim_problem,))
            z = mu + sigma * s[k]
            # evaluate fitness
            y[k] = self._evaluate_fitness(z, args)
        return s, y

    def _compute_gradients(self, s=None, y=None):
        upsilon = np.empty((self.n_individuals,))
        upsilon[np.argsort(y)] = np.copy(self._upsilon)
        g_mu = np.dot(upsilon, s)
        g_sigma = np.dot(upsilon, np.power(s, 2) - 1)
        return g_mu, g_sigma

    def _update_distribution(self, mu=None, sigma=None, g_mu=None, g_sigma=None):
        mu += self.eta_mu * sigma * g_mu
        sigma *= np.exp(self.eta_sigma / 2 * g_sigma)
        return mu, sigma

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        self.start_time = time.time()
        fitness = []  # store all fitness generated during evolution
        if fitness_function is not None:
            self.fitness_function = fitness_function
        s, mu, sigma, y = self.initialize()
        while True:
            s, y = self.iterate(s, mu, sigma, y, args)  # sample and evaluate offspring population
            if self.record_fitness:
                fitness.extend(y.tolist())
            if self._check_terminations():
                break
            g_mu, g_sigma = self._compute_gradients(s, y)
            mu, sigma = self._update_distribution(mu, sigma, g_mu, g_sigma)
            self.n_generations += 1
            self._print_verbose_info(y)
        if self.record_fitness:
            self._compress_fitness(fitness[:self.n_function_evaluations])
        results = self._collect_results()
        results['mu'] = mu
        results['sigma'] = sigma
        return results



"""Repeat the following paper for `FEP`:
    Yao, X., Liu, Y. and Lin, G., 1999.
    Evolutionary programming made faster.
    IEEE Transactions on Evolutionary Computation, 3(2), pp.82-102.
    https://ieeexplore.ieee.org/abstract/document/771163

    Since its source code is not openly available, the performance differences between our code and the original
    paper are very hard (if not impossible) to analyze. We notice that different people may give a different
    implementation based on different understandings of its algorithmic operations.

    Unfortunately, till now our code *cannot* perfectly repeat the results reported by the original paper.
    We expect that a much closer open-source implementation could be given in the future.
"""
import numpy as np

from pypop7.benchmarks.base_functions import sphere, step, rosenbrock, rastrigin, ackley
from pypop7.optimizers.ep.fep import FEP


if __name__ == '__main__':
    ndim_problem = 30

    problem = {'fitness_function': rastrigin,
               'ndim_problem': ndim_problem,
               'lower_boundary': -5.12 * np.ones((ndim_problem,)),
               'upper_boundary': 5.12 * np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 5000 * 100,
               'seed_rng': 0,  # undefined in the original paper
               'sigma': 3.0}
    fep = FEP(problem, options)
    results = fep.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 22.88404827530138
    # vs 4.6e-2 (from the original paper)

    problem = {'fitness_function': ackley,
               'ndim_problem': ndim_problem,
               'lower_boundary': -32 * np.ones((ndim_problem,)),
               'upper_boundary': 32 * np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 1500 * 100,
               'seed_rng': 0,  # undefined in the original paper
               'sigma': 3.0}
    fep = FEP(problem, options)
    results = fep.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 1.9075267778220923
    # vs 1.8e-2 (from the original paper)

    problem = {'fitness_function': sphere,
               'ndim_problem': ndim_problem,
               'lower_boundary': -100*np.ones((ndim_problem,)),
               'upper_boundary': 100*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 1500 * 100,
               'seed_rng': 0,  # undefined in the original paper
               'sigma': 3.0}
    fep = FEP(problem, options)
    results = fep.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 0.20284227582986794
    # vs 5.7e-4 (from the original paper)

    problem = {'fitness_function': step,
               'ndim_problem': ndim_problem,
               'lower_boundary': -100 * np.ones((ndim_problem,)),
               'upper_boundary': 100 * np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 1500 * 100,
               'seed_rng': 0,  # undefined in the original paper
               'sigma': 3.0}
    fep = FEP(problem, options)
    results = fep.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 47.0
    # vs 0 (from the original paper)

    problem = {'fitness_function': rosenbrock,
               'ndim_problem': ndim_problem,
               'lower_boundary': -30*np.ones((ndim_problem,)),
               'upper_boundary': 30*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 20000 * 100,
               'seed_rng': 0,  # undefined in the original paper
               'sigma': 3.0}
    fep = FEP(problem, options)
    results = fep.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 19.653232176391665
    # vs mean 5.06 std 5.87 (from the original paper)



"""Repeat the following paper for `UMDA`:
    Larranaga, P., Etxeberria, R., Lozano, J.A. and Pena, J.M., 2000.
    Optimization in continuous domains by learning and simulation of Gaussian networks.
    Technical Report, Department of Computer Science and Artificial Intelligence,
    University of the Basque Country.
    https://tinyurl.com/3bw6n3x4

    Since its source code is not openly available, the performance differences are very hard (
    if not impossible) to analyze.

    However, we again notice the following paper:
    Larrañaga, P. and Lozano, J.A. eds., 2001.
    Estimation of distribution algorithms: A new tool for evolutionary computation.
    Springer Science & Business Media.
    https://link.springer.com/book/10.1007/978-1-4615-1539-5
    (See Chapter 8 Experimental Results in Function Optimization with EDAs in Continuous Domain)

    With nearly the same settings with the first paper, very close performance can be obtained by our code.
    Therefore, we argue that the repeatability of `UMDA` can be well-documented (*at least partly*).
"""
import numpy as np

from pypop7.benchmarks.base_functions import rosenbrock, griewank
from pypop7.optimizers.eda.umda import UMDA


if __name__ == '__main__':
    ndim_problem = 10

    problem = {'fitness_function': rosenbrock,
               'ndim_problem': ndim_problem,
               'lower_boundary': -10*np.ones((ndim_problem,)),
               'upper_boundary': 10*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 300000,
               'n_individuals': 2000,
               'seed_rng': 0}  # undefined in the original paper
    umda = UMDA(problem, options)
    results = umda.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 8.221260903106916
    # vs 0.13754 (from the first original paper)
    # vs 8.7204  (from the second original paper)

    problem = {'fitness_function': griewank,
               'ndim_problem': ndim_problem,
               'lower_boundary': -600*np.ones((ndim_problem,)),
               'upper_boundary': 600*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 300000,
               'n_individuals': 750,
               'seed_rng': 0}  # undefined in the original paper
    umda = UMDA(problem, options)
    results = umda.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 0.0
    # vs 0.011076   (from the first original paper)
    # vs 6.0783e-02 (from the second original paper)

    ndim_problem = 50

    problem = {'fitness_function': rosenbrock,
               'ndim_problem': ndim_problem,
               'lower_boundary': -10 * np.ones((ndim_problem,)),
               'upper_boundary': 10 * np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 300000,
               'n_individuals': 2000,
               'seed_rng': 0}  # undefined in the original paper
    umda = UMDA(problem, options)
    results = umda.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 47.805237128504224
    # vs 48.8949  (from the second original paper)

    problem = {'fitness_function': griewank,
               'ndim_problem': ndim_problem,
               'lower_boundary': -600 * np.ones((ndim_problem,)),
               'upper_boundary': 600 * np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 300000,
               'n_individuals': 750,
               'seed_rng': 0}  # undefined in the original paper
    umda = UMDA(problem, options)
    results = umda.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 0.0
    # vs 8.9869e-06 (from the second original paper)



import numpy as np

from pypop7.optimizers.es.es import ES


class LMMAES(ES):
    """Limited-Memory Matrix Adaptation Evolution Strategy (LMMAES).

    Reference
    ---------
    Loshchilov, I., Glasmachers, T. and Beyer, H.G., 2019.
    Large scale black-box optimization by limited-memory matrix adaptation.
    IEEE Transactions on Evolutionary Computation, 23(2), pp.353-358.
    https://ieeexplore.ieee.org/abstract/document/8410043

    See the official Python version from Glasmachers:
    https://www.ini.rub.de/upload/editor/file/1604950981_dc3a4459a4160b48d51e/lmmaes.py
    """
    def __init__(self, problem, options):
        ES.__init__(self, problem, options)
        self.options = options
        n_evolution_paths = 4 + int(3 * np.log(self.ndim_problem))
        self.n_evolution_paths = options.get('n_evolution_paths', n_evolution_paths)  # m in Algorithm 1
        self.c_s = None  # c_sigma in Algorithm 1
        self._s_1 = None  # for Line 13 in Algorithm 1
        self._s_2 = None  # for Line 13 in Algorithm 1
        self._c_d = 1 / (self.ndim_problem * np.power(1.5, np.arange(self.n_evolution_paths)))
        self._c_c = None

    def initialize(self, is_restart=False):
        self.c_s = self.options.get('c_s', 2 * self.n_individuals / self.ndim_problem)
        _s_1 = 1 - self.c_s
        if _s_1 < 0:  # undefined in the original paper
            _s_1 = 0.5
        self._s_1 = _s_1
        _s_2 = self._mu_eff * self.c_s * (2 - self.c_s)
        if _s_2 < 0:  # undefined in the original paper
            _s_2 = np.power(0.5, 2)
        self._s_2 = np.sqrt(_s_2)
        self._c_c = self.n_individuals / (self.ndim_problem * np.power(4.0, np.arange(self.n_evolution_paths)))
        z = np.empty((self.n_individuals, self.ndim_problem))  # Gaussian noise for mutation
        d = np.empty((self.n_individuals, self.ndim_problem))  # search directions
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        s = np.zeros((self.ndim_problem,))  # evolution path (p in Algorithm 1)
        tm = np.zeros((self.n_evolution_paths, self.ndim_problem))  # transformation matrix M
        y = np.empty((self.n_individuals,))  # fitness (no evaluation)
        return z, d, mean, s, tm, y

    def iterate(self, z=None, d=None, mean=None, tm=None, y=None, args=None):
        for k in range(self.n_individuals):
            if self._check_terminations():
                return z, d, y
            z[k] = self.rng_optimization.standard_normal((self.ndim_problem,))
            d[k] = z[k]
            for j in range(np.minimum(self._n_generations, self.n_evolution_paths)):
                d[k] = (1 - self._c_d[j]) * d[k] + self._c_d[j] * tm[j] * np.dot(tm[j], d[k])
            y[k] = self._evaluate_fitness(mean + self.sigma * d[k], args)
        return z, d, y

    def _update_distribution(self, z=None, d=None, mean=None, s=None, tm=None, y=None):
        order = np.argsort(y)
        d_w, z_w = np.zeros((self.ndim_problem,)), np.zeros((self.ndim_problem,))
        for k in range(self.n_parents):
            d_w += self._w[k] * d[order[k]]
            z_w += self._w[k] * z[order[k]]
        # update distribution mean
        mean += self.sigma * d_w
        # update evolution path (p_c, s) and low-rank transformation matrix (tm)
        s = self._s_1 * s + self._s_2 * z_w
        for k in range(self.n_evolution_paths):  # rank-m
            _tm_1 = 1 - self._c_c[k]
            if _tm_1 < 0:  # undefined in the original paper
                _tm_1 = 0.5
            _tm_2 = self._mu_eff * self._c_c[k] * (2 - self._c_c[k])
            if _tm_2 < 0:  # undefined in the original paper
                _tm_2 = np.power(0.5, 2)
            tm[k] = _tm_1 * tm[k] + np.sqrt(_tm_2) * z_w
        # update global step-size
        self.sigma *= np.exp(self.c_s / 2 * (np.sum(np.power(s, 2)) / self.ndim_problem - 1))
        return mean, s, tm

    def restart_initialize(self, z=None, d=None, mean=None, s=None, tm=None, y=None):
        if ES.restart_initialize(self):
            z, d, mean, s, tm, y = self.initialize(True)
        return z, d, mean, s, tm, y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        z, d, mean, s, tm, y = self.initialize()
        while True:
            z, d, y = self.iterate(z, d, mean, tm, y, args)  # sample and evaluate offspring population
            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            mean, s, tm = self._update_distribution(z, d, mean, s, tm, y)
            self._n_generations += 1
            self._print_verbose_info(y)
            if self.is_restart:
                z, d, mean, s, tm, y = self.restart_initialize(z, d, mean, s, tm, y)
        results = self._collect_results(fitness, mean)
        results['s'] = s
        results['tm'] = tm
        return results



import numpy as np

from pypop7.optimizers.eda.umda import UMDA


class EMNA(UMDA):
    """Estimation of Multivariate Normal Algorithm (EMNA).

    .. note:: `EMNA` learns the *full* covariance matrix of the Gaussian sampling distribution, resulting
       in *high* time and space complexity in each generation. Therefore, it is rarely used for large-scale
       black-box optimization (LSBBO).

       It is **highly recommended** to first attempt other more advanced methods for LSBBO.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`),
                * 'record_fitness'           - flag to record fitness list to output results (`bool`, default: `False`),
                * 'record_fitness_frequency' - function evaluations frequency of recording (`int`, default: `1000`),

                  * if `record_fitness` is set to `False`, it will be ignored,
                  * if `record_fitness` is set to `True` and it is set to 1, all fitness generated during optimization
                    will be saved into output results.

                * 'verbose'                  - flag to print verbose info during optimization (`bool`, default: `True`),
                * 'verbose_frequency'        - frequency of printing verbose info (`int`, default: `10`);
              and with two particular settings (`keys`):
                * 'n_individuals' - number of offspring, offspring population size (`int`, default: `200`),
                * 'n_parents'     - number of parents, parental population size (`int`, default:
                  `int(self.n_individuals / 2)`).

    Examples
    --------
    Use the EDA optimizer `EMNA` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.eda.emna import EMNA
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5 * numpy.ones((2,)),
       ...            'upper_boundary': 5 * numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022}
       >>> emna = EMNA(problem, options)  # initialize the optimizer class
       >>> results = emna.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"UMDA: {results['n_function_evaluations']}, {results['best_so_far_y']}")
         * Generation 10: best_so_far_y 8.37514e-03, min(y) 3.38137e-01 & Evaluations 2200
         * Generation 20: best_so_far_y 8.37514e-03, min(y) 3.35835e-01 & Evaluations 4200
       UMDA: 5000, 0.008375142194038284

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/2p8xksyy>`_ for more details.

    Attributes
    ----------
    n_individuals : `int`
                    number of offspring, offspring population size.
    n_parents     : `int`
                    number of parents, parental population size.

    References
    ----------
    Larrañaga, P. and Lozano, J.A. eds., 2001.
    Estimation of distribution algorithms: A new tool for evolutionary computation.
    Springer Science & Business Media.
    https://link.springer.com/book/10.1007/978-1-4615-1539-5

    Larranaga, P., Etxeberria, R., Lozano, J.A. and Pena, J.M., 2000.
    Optimization in continuous domains by learning and simulation of Gaussian networks.
    Technical Report, Department of Computer Science and Artificial Intelligence,
    University of the Basque Country.
    https://tinyurl.com/3bw6n3x4
    """
    def __init__(self, problem, options):
        UMDA.__init__(self, problem, options)

    def iterate(self, x=None, y=None, args=None):
        order = np.argsort(y)[:self.n_parents]
        mean = np.mean(x[order], axis=0)
        cov = np.cov(np.transpose(x[order]))
        for i in range(self.n_individuals):
            if self._check_terminations():
                break
            x[i] = self.rng_optimization.multivariate_normal(mean, cov)
            y[i] = self._evaluate_fitness(x[i], args)
        return x, y



from optimizers.pso.opso import OPSO
from optimizers.pso.psort import ring_topology


class OPSORT(OPSO):
    """Online Particle Swarm Optimizer with Ring Topology (PSORT).

    Reference
    ---------
    Kennedy, J. and Eberhart, R., 1995, November.
    Particle swarm optimization.
    In Proceedings of International Conference on Neural Networks (Vol. 4, pp. 1942-1948). IEEE.
    https://ieeexplore.ieee.org/document/488968

    Shi, Y. and Eberhart, R., 1998, May.
    A modified particle swarm optimizer.
    In IEEE World Congress on Computational Intelligence (pp. 69-73). IEEE.
    https://ieeexplore.ieee.org/abstract/document/699146
    """
    def __init__(self, problems, options):
        OPSO.__init__(self, problems, options)
        self.topology = ring_topology



import numpy as np

from pypop7.optimizers.cem.cem import CEM


class SCEM(CEM):
    """Cross-Entropy Method(CEM)
    Reference
    -----------
    P. T. de Boer, D. P. Kroese, S. Mannor, R. Y. Rubinstein, (2003)
    A Tutorial on the Cross-Entropy Method
    http://web.mit.edu/6.454/www/www_fall_2003/gew/CEtutorial.pdf
    T. Hoem-de-Mello, R. Y. Rubinstein,
    Estimation of Rare Event Probabilities using Cross-Entropy
    Proceedings of the 2002 Winter Simulation Conference
    https://ieeexplore.ieee.org/stamp/stamp.jsp?tp=&arnumber=1172900
    The official python version:
    https://pypi.org/project/cross-entropy-method/
    """
    def __init__(self, problem, options):
        CEM.__init__(self, problem, options)

    def initialize(self, is_restart=False):
        mean = self._initialize_mean(is_restart)
        y = np.empty((self.n_individuals,))
        x = np.empty((self.n_individuals, self.ndim_problem))
        return mean, x, y

    def iterate(self, mean, x, y):
        for i in range(self.n_individuals):
            x[i] = self.rng_optimization.normal(mean, self.sigma, (1, self.ndim_problem))
            y[i] = self._evaluate_fitness(x[i])
            if self._check_terminations():
                return x, y
        return x, y

    def _update_parameters(self, mean, x, y):
        order = np.argsort(y)
        new_x = np.empty((self.n_parents, self.ndim_problem))
        for j in range(self.n_parents):
            new_x[j] = x[order[j]]
        mean = np.mean(new_x, axis=0)
        self.sigma = np.std(new_x, axis=0)
        return mean

    def optimize(self, fitness_function=None, args=None):
        fitness = CEM.optimize(self, fitness_function)
        mean, x, y = self.initialize()
        while True:
            x, y = self.iterate(mean, x, y)
            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            self._n_generations += 1
            self._print_verbose_info(y)
            mean = self._update_parameters(mean, x, y)
        results = self._collect_results(fitness, mean)
        return results



import numpy as np

from optimizers.pso.opso import OPSO


class CLPSO(OPSO):
    """Comprehensive Learning Particle Swarm Optimizer (CLPSO).

    Reference
    ---------
    Liang, J.J., Qin, A.K., Suganthan, P.N. and Baskar, S., 2006.
    Comprehensive learning particle swarm optimizer for global optimization of multimodal functions.
    IEEE Transactions on Evolutionary Computation, 10(3), pp.281-295.
    https://ieeexplore.ieee.org/abstract/document/1637688

    Original MATLAB Source Code:
    https://www3.ntu.edu.sg/home/epnsugan/
    https://github.com/P-N-Suganthan/CODES/blob/master/2006-IEEE-TEC-CLPSO.zip
    """
    def __init__(self, problems, options):
        OPSO.__init__(self, problems, options)
        # comprehensive-learning rate including cognition-learning and society-learning
        self.c = options.get('c', 1.49445)
        # refreshing gap (m), same with the original code (5) but different from the original paper (7)
        self.m = options.get('m', 5)
        # learning rates for all particles, same with the original code but different from the original paper
        pc = 5 * np.hstack((np.arange(0, 1, 1 / (self.n_individuals - 1)), 1))
        self.Pc = 0.5 * (np.exp(pc) - np.exp(pc[0])) / (np.exp(pc[-1]) - np.exp(pc[0]))
        # number of successive generations each particle has not improved its own personally previous-best fitness
        self.flag = np.zeros((self.n_individuals,))
        # linearly decreasing inertia weight, same with the original code but different from the original paper
        self.w = 0.9 - np.arange(1, self.max_generations + 1) * (0.7 / self.max_generations)
        self.topology = self.learning_topology

    def learning_topology(self, p_x, p_y, i, n_x):
        # for clearer coding, it is slightly different from the original code
        # which used the vector-based form (hard to read), but it is same with the original paper
        if self.flag[i] >= self.m:
            self.flag[i] = 0
            all_exemplars = i * np.ones((self.ndim_problem,))
            for d in range(self.ndim_problem):
                if self.rng_optimization.random() < self.Pc[i]:  # tournament selection
                    left, right = self.rng_optimization.choice(self.n_individuals, 2, replace=False)
                    if p_y[left] < p_y[right]:
                        n_x[i, d] = p_x[left, d]
                        all_exemplars[d] = left
                    else:
                        n_x[i, d] = p_x[right, d]
                        all_exemplars[d] = right
                else:
                    n_x[i, d] = p_x[i, d]
            if np.alltrue(all_exemplars == i):
                possible_exemplars = set(range(self.n_individuals))
                possible_exemplars.remove(i)
                exemplar_dim = self.rng_optimization.integers(self.ndim_problem)
                exemplar = self.rng_optimization.choice(list(possible_exemplars))
                n_x[i, exemplar_dim] = p_x[exemplar, exemplar_dim]
        return n_x[i], i

    def iterate(self, x=None, y=None, p_x=None, p_y=None, n_x=None, v=None, args=None):
        for i in range(self.n_individuals):  # use online (rather batch) update
            if self._check_terminations():
                return x, y, p_x, p_y, n_x, v
            # update neighbor topology
            n_x[i], _ = self.topology(p_x, p_y, i, n_x)
            # update and limit velocity
            learning_rand = self.rng_optimization.uniform(size=(self.ndim_problem,))
            v[i] = self.w[self.n_generations] * v[i] + self.c * learning_rand * (n_x[i] - x[i])
            less_min_v, more_max_v = v[i] < self.min_v, v[i] > self.max_v
            v[i, less_min_v], v[i, more_max_v] = self.min_v[less_min_v], self.max_v[more_max_v]
            # update position
            x[i] += v[i]
            # evaluate fitness
            y[i] = self._evaluate_fitness(x[i], args)
            if y[i] < p_y[i]:
                p_x[i], p_y[i] = x[i], y[i]
            else:
                self.flag[i] += 1
        return x, y, p_x, p_y, n_x, v



import os
import numpy as np

from pypop7.benchmarks import base_functions
from pypop7.benchmarks.base_functions import squeeze_and_check, BaseFunction


# helper functions
def generate_shift_vector(func, ndim, low, high, seed=None):
    """Generate a random shift vector of dimension `ndim`, sampled uniformly between
        `low` (inclusive) and `high` (exclusive).

    Note that the generated shift vector will be automatically stored in txt form for further use.

    :param func: function name, a `str` or `function` object.
    :param ndim: number of dimensions of the shift vector, an `int` scalar.
    :param low: lower boundary of the shift vector, a `float` scalar or array_like of floats.
    :param high: upper boundary of the shift vector, a `float` scalar or array_like of floats.
    :param seed: seed for random number generator, a `int` scalar.
    :return: shift vector, a `ndim`-d vector sampled uniformly in [`low`, `high`).

    https://numpy.org/doc/stable/reference/random/generated/numpy.random.Generator.uniform.html
    """
    low, high = squeeze_and_check(low), squeeze_and_check(high)
    if hasattr(func, '__call__'):
        func = func.__name__
    data_folder = 'pypop_benchmarks_input_data'
    if not os.path.exists(data_folder):
        os.mkdir(data_folder)
    data_path = os.path.join(data_folder, 'shift_vector_' + func + '_dim_' + str(ndim) + '.txt')
    shift_vector = np.random.default_rng(seed).uniform(low, high, size=ndim)
    np.savetxt(data_path, shift_vector)
    return shift_vector


def load_shift_vector(func, x, shift_vector=None):
    """Load the shift vector which needs to be generated in advance.
        When `None`, the shift vector should have been generated and stored in txt form in advance.

    :param func: function name, a `function` object.
    :param x: decision vector, array_like of floats.
    :param shift_vector: shift vector, array_like of floats.
    :return: shift vector, a 1-d `ndarray` of `dtype` `np.float64` with the same size as `x`.
    """
    x = squeeze_and_check(x)
    if shift_vector is None:
        if (not hasattr(func, 'pypop_shift_vector')) or (func.pypop_shift_vector.size != x.size):
            data_folder = 'pypop_benchmarks_input_data'
            data_path = os.path.join(data_folder, 'shift_vector_' + func.__name__ + '_dim_' + str(x.size) + '.txt')
            shift_vector = np.loadtxt(data_path)
            func.pypop_shift_vector = shift_vector
        shift_vector = func.pypop_shift_vector
    shift_vector = squeeze_and_check(shift_vector)
    if shift_vector.shape != x.shape:
        raise TypeError('shift_vector should have the same shape as x.')
    if shift_vector.size != x.size:
        raise TypeError('shift_vector should have the same size as x.')
    return shift_vector


def sphere(x, shift_vector=None):
    shift_vector = load_shift_vector(sphere, x, shift_vector)
    y = base_functions.sphere(x - shift_vector)
    return y


class Sphere(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'sphere'

    def __call__(self, x, shift_vector=None):
        return sphere(x, shift_vector)


def cigar(x, shift_vector=None):
    shift_vector = load_shift_vector(cigar, x, shift_vector)
    y = base_functions.cigar(x - shift_vector)
    return y


class Cigar(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'cigar'

    def __call__(self, x, shift_vector=None):
        return cigar(x, shift_vector)


def discus(x, shift_vector=None):  # also called tablet
    shift_vector = load_shift_vector(discus, x, shift_vector)
    y = base_functions.discus(x - shift_vector)
    return y


class Discus(BaseFunction):  # also called Tablet
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'discus'

    def __call__(self, x, shift_vector=None):
        return discus(x, shift_vector)


def cigar_discus(x, shift_vector=None):
    shift_vector = load_shift_vector(cigar_discus, x, shift_vector)
    y = base_functions.cigar_discus(x - shift_vector)
    return y


class CigarDiscus(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'cigar_discus'

    def __call__(self, x, shift_vector=None):
        return cigar_discus(x, shift_vector)


def ellipsoid(x, shift_vector=None):
    shift_vector = load_shift_vector(ellipsoid, x, shift_vector)
    y = base_functions.ellipsoid(x - shift_vector)
    return y


class Ellipsoid(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'ellipsoid'

    def __call__(self, x, shift_vector=None):
        return ellipsoid(x, shift_vector)


def different_powers(x, shift_vector=None):
    shift_vector = load_shift_vector(different_powers, x, shift_vector)
    y = base_functions.different_powers(x - shift_vector)
    return y


class DifferentPowers(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'different_powers'

    def __call__(self, x, shift_vector=None):
        return different_powers(x, shift_vector)


def schwefel221(x, shift_vector=None):
    shift_vector = load_shift_vector(schwefel221, x, shift_vector)
    y = base_functions.schwefel221(x - shift_vector)
    return y


class Schwefel221(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'schwefel221'

    def __call__(self, x, shift_vector=None):
        return schwefel221(x, shift_vector)


def step(x, shift_vector=None):
    shift_vector = load_shift_vector(step, x, shift_vector)
    y = base_functions.step(x - shift_vector)
    return y


class Step(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'step'

    def __call__(self, x, shift_vector=None):
        return step(x, shift_vector)


def schwefel222(x, shift_vector=None):
    shift_vector = load_shift_vector(schwefel222, x, shift_vector)
    y = base_functions.schwefel222(x - shift_vector)
    return y


class Schwefel222(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'schwefel222'

    def __call__(self, x, shift_vector=None):
        return schwefel222(x, shift_vector)


def rosenbrock(x, shift_vector=None):
    shift_vector = load_shift_vector(rosenbrock, x, shift_vector)
    y = base_functions.rosenbrock(x - shift_vector)
    return y


class Rosenbrock(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'rosenbrock'

    def __call__(self, x, shift_vector=None):
        return rosenbrock(x, shift_vector)


def schwefel12(x, shift_vector=None):
    shift_vector = load_shift_vector(schwefel12, x, shift_vector)
    y = base_functions.schwefel12(x - shift_vector)
    return y


class Schwefel12(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'schwefel12'

    def __call__(self, x, shift_vector=None):
        return schwefel12(x, shift_vector)


def exponential(x, shift_vector=None):
    shift_vector = load_shift_vector(exponential, x, shift_vector)
    y = base_functions.exponential(x - shift_vector)
    return y


class Exponential(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'exponential'

    def __call__(self, x, shift_vector=None):
        return exponential(x, shift_vector)


def griewank(x, shift_vector=None):
    shift_vector = load_shift_vector(griewank, x, shift_vector)
    y = base_functions.griewank(x - shift_vector)
    return y


class Griewank(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'griewank'

    def __call__(self, x, shift_vector=None):
        return griewank(x, shift_vector)


def bohachevsky(x, shift_vector=None):
    shift_vector = load_shift_vector(bohachevsky, x, shift_vector)
    y = base_functions.bohachevsky(x - shift_vector)
    return y


class Bohachevsky(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'bohachevsky'

    def __call__(self, x, shift_vector=None):
        return bohachevsky(x, shift_vector)


def ackley(x, shift_vector=None):
    shift_vector = load_shift_vector(ackley, x, shift_vector)
    y = base_functions.ackley(x - shift_vector)
    return y


class Ackley(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'ackley'

    def __call__(self, x, shift_vector=None):
        return ackley(x, shift_vector)


def rastrigin(x, shift_vector=None):
    shift_vector = load_shift_vector(rastrigin, x, shift_vector)
    y = base_functions.rastrigin(x - shift_vector)
    return y


class Rastrigin(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'rastrigin'

    def __call__(self, x, shift_vector=None):
        return rastrigin(x, shift_vector)


def scaled_rastrigin(x, shift_vector=None):
    shift_vector = load_shift_vector(scaled_rastrigin, x, shift_vector)
    y = base_functions.scaled_rastrigin(x - shift_vector)
    return y


class ScaledRastrigin(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'scaled_rastrigin'

    def __call__(self, x, shift_vector=None):
        return scaled_rastrigin(x, shift_vector)


def skew_rastrigin(x, shift_vector=None):
    shift_vector = load_shift_vector(skew_rastrigin, x, shift_vector)
    y = base_functions.skew_rastrigin(x - shift_vector)
    return y


class SkewRastrigin(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'skew_rastrigin'

    def __call__(self, x, shift_vector=None):
        return skew_rastrigin(x, shift_vector)


def levy_montalvo(x, shift_vector=None):
    shift_vector = load_shift_vector(levy_montalvo, x, shift_vector)
    y = base_functions.levy_montalvo(x - shift_vector)
    return y


class LevyMontalvo(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'levy_montalvo'

    def __call__(self, x, shift_vector=None):
        return levy_montalvo(x, shift_vector)


def michalewicz(x, shift_vector=None):
    shift_vector = load_shift_vector(michalewicz, x, shift_vector)
    y = base_functions.michalewicz(x - shift_vector)
    return y


class Michalewicz(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'michalewicz'

    def __call__(self, x, shift_vector=None):
        return michalewicz(x, shift_vector)


def salomon(x, shift_vector=None):
    shift_vector = load_shift_vector(salomon, x, shift_vector)
    y = base_functions.salomon(x - shift_vector)
    return y


class Salomon(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'salomon'

    def __call__(self, x, shift_vector=None):
        return salomon(x, shift_vector)


def shubert(x, shift_vector=None):
    shift_vector = load_shift_vector(shubert, x, shift_vector)
    y = base_functions.shubert(x - shift_vector)
    return y


class Shubert(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'shubert'

    def __call__(self, x, shift_vector=None):
        return shubert(x, shift_vector)


def schaffer(x, shift_vector=None):
    shift_vector = load_shift_vector(schaffer, x, shift_vector)
    y = base_functions.schaffer(x - shift_vector)
    return y


class Schaffer(BaseFunction):
    def __init__(self):
        BaseFunction.__init__(self)
        self.__name__ = 'schaffer'

    def __call__(self, x, shift_vector=None):
        return schaffer(x, shift_vector)



from pypop7.optimizers.es.maes import MAES


class FMAES(MAES):
    """Fast Matrix Adaptation Evolution Strategy (FMAES).

    .. note:: `FMAES` is a more efficient implementation of `MAES` with *quadractic* time complexity.

       It is **highly recommended** to first attempt other more advanced ES variants (e.g., `LM-CMA`, `LM-MA-ES`) for
       large-scale black-box optimization (LSBBO), since `FMAES` has a *quadratic* time complexity w.r.t. each sampling.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`),
                * 'record_fitness'           - flag to record fitness list to output results (`bool`, default: `False`),
                * 'record_fitness_frequency' - function evaluations frequency of recording (`int`, default: `1000`),

                  * if `record_fitness` is set to `False`, it will be ignored,
                  * if `record_fitness` is set to `True` and it is set to 1, all fitness generated during optimization
                    will be saved into output results.

                * 'verbose'                  - flag to print verbose info during optimization (`bool`, default: `True`),
                * 'verbose_frequency'        - generation frequency of printing verbose info (`int`, default: `10`);
              and with four particular settings (`keys`):
                * 'sigma'         - initial global step-size (σ), mutation strength (`float`),
                * 'mean'          - initial (starting) point, mean of Gaussian search distribution (`array_like`),

                  * if not given, it will draw a random sample from the uniform distribution whose search range is
                    bounded by `problem['lower_boundary']` and `problem['upper_boundary']`).

                * 'n_individuals' - number of offspring (λ: lambda), offspring population size (`int`, default:
                  `4 + int(3*np.log(self.ndim_problem))`),
                * 'n_parents'     - number of parents (μ: mu), parental population size (`int`, default:
                  `int(self.n_individuals / 2)`).

    Examples
    --------
    Use the ES optimizer `FMAES` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.es.fmaes import FMAES
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5 * numpy.ones((2,)),
       ...            'upper_boundary': 5 * numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'mean': 3 * numpy.ones((2,)),
       ...            'sigma': 0.1}
       >>> fmaes = FMAES(problem, options)  # initialize the optimizer class
       >>> results = fmaes.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"FMAES: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       FMAES: 5000, 1.211565542464775e-18

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/37ews6h4>`_ for more details.

    Attributes
    ----------
    n_individuals   : `int`
                      number of offspring (λ: lambda), offspring population size.
    n_parents       : `int`
                      number of parents (μ: mu), parental population size.
    mean            : `array_like`
                      initial (starting) point, mean of Gaussian search distribution.
    sigma           : `float`
                      initial global step-size (σ), mutation strength.

    References
    ----------
    Beyer, H.G., 2020, July.
    Design principles for matrix adaptation evolution strategies.
    In Proceedings of Annual Conference on Genetic and Evolutionary Computation Companion (pp. 682-700).
    https://dl.acm.org/doi/abs/10.1145/3377929.3389870

    Loshchilov, I., Glasmachers, T. and Beyer, H.G., 2019.
    Large scale black-box optimization by limited-memory matrix adaptation.
    IEEE Transactions on Evolutionary Computation, 23(2), pp.353-358.
    https://ieeexplore.ieee.org/abstract/document/8410043

    Beyer, H.G. and Sendhoff, B., 2017.
    Simplify your covariance matrix adaptation evolution strategy.
    IEEE Transactions on Evolutionary Computation, 21(5), pp.746-759.
    https://ieeexplore.ieee.org/document/7875115
    """
    def __init__(self, problem, options):
        options['_fast_version'] = True  # mandatory setting for FMAES
        MAES.__init__(self, problem, options)



"""Repeat Fig. 3 (a) and (d) from the following paper:
    Li, Z. and Zhang, Q., 2018.
    A simple yet efficient evolution strategy for large-scale black-box optimization.
    IEEE Transactions on Evolutionary Computation, 22(5), pp.637-646.
    https://ieeexplore.ieee.org/abstract/document/8080257
"""
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

from optimizers.es.rmes2 import RMES2

sns.set_theme(style='darkgrid')


if __name__ == '__main__':
    # plot Fig. 3(a)
    from pypop7.benchmarks.base_functions import ellipsoid
    ndim_problem = 1000
    problem = {'fitness_function': ellipsoid,
               'ndim_problem': ndim_problem,
               'lower_boundary': -10 * np.ones((ndim_problem,)),
               'upper_boundary': 10 * np.ones((ndim_problem,))}
    options = {'fitness_threshold': 1e-8,
               'max_function_evaluations': 1e8,
               'seed_rng': 2022,  # not given in the original paper
               'sigma': 20 / 3,
               'verbose_frequency': 20000,
               'record_fitness': True,
               'record_fitness_frequency': 1}
    rmes = RMES2(problem, options)
    results = rmes.optimize()
    plt.figure()
    plt.xscale('log')
    plt.yscale('log')
    plt.yticks(ticks=[1e-8, 1e-4, 1e0, 1e4, 1e8])
    plt.xticks(ticks=[1e2, 1e4, 1e6, 1e8])
    plt.xlim([1e2, 1e8])
    fitness = results['fitness']
    plt.plot(fitness[:, 0], fitness[:, 1], 'r-')
    plt.xlabel('Function Evaluations')
    plt.ylabel('Objective Value')
    plt.show()
    # plot Fig. 3(d)
    from pypop7.benchmarks.base_functions import rosenbrock
    ndim_problem = 1000
    problem = {'fitness_function': rosenbrock,
               'ndim_problem': ndim_problem,
               'lower_boundary': -10 * np.ones((ndim_problem,)),
               'upper_boundary': 10 * np.ones((ndim_problem,))}
    options = {'fitness_threshold': 1e-8,
               'max_function_evaluations': 1e8,
               'seed_rng': 2022,  # not given in the original paper
               'sigma': 20 / 3,
               'verbose_frequency': 20000,
               'record_fitness': True,
               'record_fitness_frequency': 1}
    rmes = RMES2(problem, options)
    results = rmes.optimize()
    plt.figure()
    plt.xscale('log')
    plt.yscale('log')
    plt.yticks(ticks=[1e-8, 1e-4, 1e0, 1e4, 1e8])
    plt.xticks(ticks=[1e2, 1e4, 1e6, 1e8])
    plt.xlim([1e2, 1e8])
    fitness = results['fitness']
    plt.plot(fitness[:, 0], fitness[:, 1], 'r-')
    plt.xlabel('Function Evaluations')
    plt.ylabel('Objective Value')
    plt.show()



import time
import numpy as np

from optimizers.core.optimizer import Optimizer


class PSO(Optimizer):
    """Particle Swarm Optimizer (PSO).

    Reference
    ---------
    Kennedy, J. and Eberhart, R., 1995, November.
    Particle swarm optimization.
    In Proceedings of International Conference on Neural Networks (Vol. 4, pp. 1942-1948). IEEE.
    https://ieeexplore.ieee.org/document/488968

    Shi, Y. and Eberhart, R., 1998, May.
    A modified particle swarm optimizer.
    In IEEE World Congress on Computational Intelligence (pp. 69-73). IEEE.
    https://ieeexplore.ieee.org/abstract/document/699146

    https://github.com/pybrain/pybrain/blob/master/pybrain/optimization/populationbased/pso.py
    """
    def __init__(self, problem, options):
        Optimizer.__init__(self, problem, options)
        if self.n_individuals is None:  # swarm (population) size
            self.n_individuals = 20  # number of particles
        self.swarm_size = (self.n_individuals, self.ndim_problem)
        self.cognition = options.get('cognition', 2.0)  # cognition-learning rate
        self.society = options.get('society', 2.0)  # society-learning rate
        self.topology = None  # to control neighbors of society learning
        self.max_ratio_v = options.get('max_ratio_v', 0.2)  # maximal ratio of velocities w.r.t. entire search range
        self.max_v = self.max_ratio_v * (self.upper_boundary - self.lower_boundary)
        self.min_v = -self.max_v
        self.n_generations = options.get('n_generations', 0)
        self.max_generations = int(np.ceil(self.max_function_evaluations / self.n_individuals))
        w = 0.9 - np.arange(1, self.max_generations + 1) * 0.5 / self.max_generations
        self.w = options.get('w', w)  # inertia weight
        self.record_fitness_initialization = False  # record all fitness generated during initialization

    def initialize(self, args=None):
        rng = self.rng_initialization
        x = rng.uniform(self.initial_lower_boundary, self.initial_upper_boundary, size=self.swarm_size)  # positions
        y = np.empty((self.n_individuals,))  # fitness
        p_x, p_y = np.copy(x), np.copy(y)  # personally previous-best positions and fitness
        n_x = np.copy(x)  # neighborly previous-best positions
        v = rng.uniform(self.min_v, self.max_v, size=self.swarm_size)  # velocities
        return x, y, p_x, p_y, n_x, v

    def iterate(self, x=None, y=None, p_x=None, p_y=None, n_x=None, v=None, args=None):  # use batch update
        rng = self.rng_optimization
        # evaluate fitness
        for i in range(self.n_individuals):
            if self._check_terminations():
                return x, y, p_x, p_y, n_x, v
            y[i] = self._evaluate_fitness(x[i], args)
            if y[i] < p_y[i]:
                p_x[i], p_y[i] = np.copy(x[i]), y[i]
        # update neighbor topology of each particle
        for i in range(self.n_individuals):
            n_x[i], _ = self.topology(p_x, p_y, i)
        # update and limit velocities of particles
        cognition_rand, society_rand = rng.uniform(size=self.swarm_size), rng.uniform(size=self.swarm_size)
        v = self.w[self.n_generations] * v +\
            self.cognition * cognition_rand * (p_x - x) +\
            self.society * society_rand * (n_x - x)
        for i in range(self.n_individuals):
            less_min_v, more_max_v = v[i] < self.min_v, v[i] > self.max_v
            v[i, less_min_v] = self.min_v[less_min_v]
            v[i, more_max_v] = self.max_v[more_max_v]
        # update positions of particles
        x += v
        return x, y, p_x, p_y, n_x, v

    def optimize(self, fitness_function=None, args=None):
        self.start_time = time.time()
        fitness = []  # store all fitness generated during evolution
        if fitness_function is not None:
            self.fitness_function = fitness_function
        x, y, p_x, p_y, n_x, v = self.initialize(args)
        if self.record_fitness_initialization and self.record_options['record_fitness']:
            fitness.extend(y)
        while True:
            x, y, p_x, p_y, n_x, v = self.iterate(x, y, p_x, p_y, n_x, v, args)
            if self.record_options['record_fitness']:
                fitness.extend(y)
            if self._check_terminations():
                break
            self.n_generations += 1
            self._print_verbose_info(y)
        if self.record_options['record_fitness']:
            self._compress_fitness(fitness[:self.n_function_evaluations])
        return self._collect_results()

    def _print_verbose_info(self, y=None):
        if self.verbose_options['verbose']:
            if not self.n_generations % self.verbose_options['frequency_verbose']:
                info = '  * Generation {:d}: best_so_far_y {:7.5e}, min(y) {:7.5e} & Evaluations {:d}'
                print(info.format(self.n_generations, self.best_so_far_y, np.min(y), self.n_function_evaluations))



import numpy as np

from pypop7.optimizers.es.es import ES


class DES(ES):
    """Differential Evolution Strategy (DES)

    Reference
    ---------
    Arabas, J., and Jagodziński, D. 2020.
    Toward a matrix-free covariance matrix adaptation evolution strategy.
    IEEE Transactions on Evolutionary Computation, 24(1), 84–98.
    https://doi.org/10.1109/TEVC.2019.2907266
    """
    def __init__(self, problem, options):
        ES.__init__(self, problem, options)
        self.c_d = options.get('c_d', self.ndim_problem / (self.ndim_problem + 2))  # for Line 16, 17, 18 in Fig. 4.
        self.c = options.get('c', 1 / np.sqrt(self.ndim_problem))  # for Line 11 in Fig. 4. (c_c)
        self.h = options.get('h', int(6 + 3 * np.sqrt(self.ndim_problem)))  # window size, for Line 14 in Fig. 4. (H)
        self.c_cov = options.get('c_cov', 2 / (self.ndim_problem ** 2))  # for Line 19 in Fig. 4. (c_epsilon)
        self.epsilon = 1e-6  # for Line 19 in Fig. 4.

    def initialize(self, args=None):
        x = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary,
                                            size=(self.n_individuals, self.ndim_problem))
        y = np.empty((self.n_individuals,))
        for i in range(self.n_individuals):
            if self._check_terminations():
                break
            y[i] = self._evaluate_fitness(x[i], args)
        mean = np.mean(x, 0)
        accu_p = np.empty((0, self.ndim_problem))
        accu_d = np.empty((0, self.ndim_problem))
        accu_x = np.empty((0, self.ndim_problem))
        self._n_generations += 1
        return x, mean, accu_p, accu_d, accu_x, y

    def iterate(self, x=None, mean=None, accu_p=None, accu_d=None, accu_x=None, y=None, args=None):
        for k in range(self.n_individuals):  # for Line 13 in Fig. 4.
            if self._check_terminations():
                return x, y
            tau = self.rng_optimization.choice(np.min([self._n_generations, self.h]), (3,))  # for Line 14 in Fig. 4.
            x1, x2 = self.rng_optimization.choice(
                accu_x[(tau[0] * self.n_parents):((tau[0] + 1) * self.n_parents)], (2,))
            d = np.sqrt(self.c_d / 2) * (x1 - x2) +\
                np.sqrt(self.c_d) * accu_d[tau[1]] * self.rng_optimization.standard_normal() +\
                np.sqrt(1 - self.c_d) * accu_p[tau[2]] * self.rng_optimization.standard_normal() +\
                self.epsilon * np.power((1 - self.c_cov), self._n_generations / 2) *\
                self.rng_optimization.standard_normal((self.ndim_problem,))
            x[k] = mean + d
            y[k] = self._evaluate_fitness(x[k], args)
        return x, y

    def _update_distribution(self, x=None, mean=None, accu_p=None, accu_d=None, accu_x=None, y=None):
        mean_bak = np.copy(mean)
        order = np.argsort(y)[:self.n_parents]
        mean = np.mean(x[order], 0)
        d = mean - mean_bak  # for Line 7 in Fig. 4.
        if self._n_generations == 1:
            p = d  # for Line 9 in Fig. 4.
        else:
            p = (1 - self.c) * accu_p[-1] + np.sqrt(self.n_parents * self.c * (2 - self.c)) * d
        accu_p = np.vstack((accu_p, p))[-self.h:]
        accu_d = np.vstack((accu_d, d))[-self.h:]
        accu_x = np.vstack((accu_x, x[order]))[-self.h * self.n_parents:]
        return mean, accu_p, accu_d, accu_x

    def optimize(self, fitness_function=None, args=None):
        fitness = ES.optimize(self, fitness_function)
        x, mean, accu_p, accu_d, accu_x, y = self.initialize()
        fitness.extend(y)
        while True:
            mean, accu_p, accu_d, accu_x = self._update_distribution(x, mean, accu_p, accu_d, accu_x, y)
            x, y = self.iterate(x, mean, accu_p, accu_d, accu_x, y, args)
            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            self._n_generations += 1
            self._print_verbose_info(y)
        results = self._collect_results(fitness)
        return results



import numpy as np
from scipy.stats import norm

from pypop7.optimizers.es.es import ES


class MMES(ES):
    """Mixture Model-based Evolution Strategy (MMES).
    Reference
    ---------
    He, X., Zheng, Z. and Zhou, Y., 2021.
    MMES: Mixture model-based evolution strategy for large-scale optimization.
    IEEE Transactions on Evolutionary Computation, 25(2), pp.320-333.
    https://ieeexplore.ieee.org/abstract/document/9244595
    See the official Matlab version from He:
    https://github.com/hxyokokok/MMES
    """
    def __init__(self, problem, options):
        ES.__init__(self, problem, options)
        # number of candidate direction vectors
        self.m = options.get('m', 2 * int(np.ceil(np.sqrt(self.ndim_problem))))
        # learning rate of evolution path
        self.c_c = options.get('c_c', 0.4 / np.sqrt(self.ndim_problem))
        self.ms = options.get('ms', 4)  # mixing strength (l)
        # for paired test adaptation (PTA)
        self.c_s = options.get('c_s', 0.3)  # learning rate of step-size adaptation
        self.a_z = options.get('a_z', 0.05)  # target significance level
        # minimal distance of updating evolution paths (T)
        self.distance = options.get('distance', np.ceil(1 / self.c_c))
        # success probability of geometric distribution (different from 4/n in the original paper)
        self.c_a = options.get('c_a', 3.8 / self.ndim_problem)  # same as Matlab code
        self.gamma = options.get('gamma', 1 - np.power(1 - self.c_a, self.m))
        self._n_mirror_sampling = None
        self._z_1 = np.sqrt(1 - self.gamma)
        self._z_2 = np.sqrt(self.gamma / self.ms)
        self._p_1 = 1 - self.c_c
        self._p_2 = np.sqrt(self.c_c * (2 - self.c_c))
        self._w_1 = 1 - self.c_s
        self._w_2 = np.sqrt(self.c_s * (2 - self.c_s))

    def initialize(self, args=None, is_restart=False):
        self._n_mirror_sampling = int(np.ceil(self.n_individuals / 2))
        x = np.zeros((self.n_individuals, self.ndim_problem))  # offspring population
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        p = np.zeros((self.ndim_problem,))  # evolution path (Line 2 in Algorithm 1)
        w = 0  # Line 3
        q = np.zeros((self.m, self.ndim_problem))  # candidate direction vectors (Line 4)
        t = np.zeros((self.m,))  # Line 5 (generations recorded)
        v = np.arange(self.m)  # Line 6 (indexes to evolution paths)
        y = np.tile(self._evaluate_fitness(mean, args), (self.n_individuals,))  # fitness
        return x, mean, p, w, q, t, v, y

    def iterate(self, x=None, mean=None, q=None, v=None, y=None, args=None):
        for k in range(self._n_mirror_sampling):  # mirror sampling
            zq = np.zeros((self.ndim_problem,))
            for _ in range(self.ms):
                j_k = v[(self.m - self.rng_optimization.geometric(self.c_a) % self.m) - 1]
                zq += self.rng_optimization.standard_normal() * q[j_k]
            z = self._z_1 * self.rng_optimization.standard_normal((self.ndim_problem,))
            z += self._z_2 * zq  # Line 13
            x[k] = mean + self.sigma * z  # Line 14
            if (self._n_mirror_sampling + k) < self.n_individuals:
                x[self._n_mirror_sampling + k] = mean - self.sigma * z
        for k in range(self.n_individuals):
            if self._check_terminations():
                return x, y
            y[k] = self._evaluate_fitness(x[k], args)
        return x, y

    def _update_distribution(self, x, mean, p, w, q, t, v, y, y_bak):
        order = np.argsort(y)
        y.sort()  # Line 16
        mean_w = np.zeros((self.ndim_problem,))
        for k in range(self.n_parents):  # Line 17
            mean_w += self._w[k] * x[order[k]]
        p = self._p_1 * p + self._p_2 * np.sqrt(self._mu_eff) * (mean_w - mean) / self.sigma  # Line 18
        mean = mean_w  # for Line 17
        if self._n_generations < self.m:
            q[self._n_generations] = p
        else:
            k_star = np.argmin(t[v[1:]] - t[v[:(self.m - 1)]])  # Line 19
            k_star += 1
            if t[v[k_star]] - t[v[k_star - 1]] > self.distance:  # Line 20
                k_star = 0  # Line 21
            v = np.append(np.append(v[:k_star], v[(k_star + 1):]), v[k_star])
            t[v[-1]], q[v[-1]] = self._n_generations, p  # Line 25, 26
        # for success-based adaptation of mutation strength
        l_w = np.dot(self._w, y_bak[:self.n_parents] > y[:self.n_parents])  # Line 27
        w = self._w_1 * w + self._w_2 * np.sqrt(self._mu_eff) * (2 * l_w - 1)  # Line 28
        self.sigma *= np.exp(norm.cdf(w) - 1 + self.a_z)  # Line 29
        return mean, p, w, q, t, v

    def restart_initialize(self, args=None, x=None, mean=None, p=None, w=None, q=None,
                           t=None, v=None, y=None, fitness=None):
        is_restart = ES.restart_initialize(self)
        if is_restart:
            x, mean, p, w, q, t, v, y = self.initialize(args, is_restart)
            fitness.append(y[0])
        return x, mean, p, w, q, t, v, y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        x, mean, p, w, q, t, v, y = self.initialize(args)
        fitness.append(y[0])
        while True:
            y_bak = np.copy(y)
            x, y = self.iterate(x, mean, q, v, y, args)  # sample and evaluate offspring population
            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            mean, p, w, q, t, v = self._update_distribution(x, mean, p, w, q, t, v, y, y_bak)
            self._n_generations += 1
            self._print_verbose_info(y)
            if self.is_restart:
                x, mean, p, w, q, t, v, y = self.restart_initialize(
                    args, x, mean, p, w, q, t, v, y, fitness)
        results = self._collect_results(fitness, mean)
        results['p'] = p
        results['w'] = w
        return results



"""Repeat Fig. 3 (Ellipsoid, Discus, Rosenbrock) from the following paper:
    Li, Z. and Zhang, Q., 2018.
    A simple yet efficient evolution strategy for large-scale black-box optimization.
    IEEE Transactions on Evolutionary Computation, 22(5), pp.637-646.
    https://ieeexplore.ieee.org/abstract/document/8080257
"""
import pickle
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

from pypop7.benchmarks.base_functions import ellipsoid, discus, rosenbrock
from pypop7.optimizers.es.rmes import RMES

sns.set_theme(style='darkgrid')


def read_pickle(function, ndim):
    file = function + '_' + str(ndim) + '.pickle'
    with open(file, 'rb') as handle:
        result = pickle.load(handle)
        return result


def write_pickle(function, ndim, result):
    file = open(function + '_' + str(ndim) + '.pickle', 'wb')
    pickle.dump(result, file)
    file.close()


def plot(function, ndim):
    plt.figure()
    result = read_pickle(function, ndim)
    plt.yscale('log')
    plt.xscale('log')
    plt.plot(result['fitness'][:, 0], result['fitness'][:, 1], color='orange')
    plt.xlim([1e2, 1e8])
    plt.xticks([1e2, 1e4, 1e6, 1e8])
    if function == 'ellipsoid':
        plt.ylim([1e-8, 1e10])
        plt.yticks(ticks=[1e-8, 1e-4, 1e0, 1e4, 1e8])
    elif function == 'discus':
        plt.ylim([1e-8, 1e6])
        plt.yticks(ticks=[1e-8, 1e-4, 1e0, 1e4])
    elif function == 'rosenbrock':
        plt.ylim([1e-8, 1e10])
        plt.yticks(ticks=[1e-8, 1e-4, 1e0, 1e4, 1e8])
    plt.xlabel("Function Evaluations")
    plt.ylabel("Objective Value")
    plt.title(function.capitalize())
    plt.show()


if __name__ == '__main__':
    ndim_problem = 1000
    for f in [ellipsoid, discus, rosenbrock]:
        problem = {'fitness_function': f,
                   'ndim_problem': ndim_problem,
                   'lower_boundary': -10 * np.ones((ndim_problem,)),
                   'upper_boundary': 10 * np.ones((ndim_problem,))}
        options = {'fitness_threshold': 1e-8,
                   'max_function_evaluations': 1e8,
                   'seed_rng': 2022,  # not given in the original paper
                   'sigma': 20 / 3,
                   'verbose_frequency': 20000,
                   'record_fitness': True,
                   'record_fitness_frequency': 1}
        rmes = RMES(problem, options)
        results = rmes.optimize()
        write_pickle(f.__name__, ndim_problem, results)
        plot(f.__name__, ndim_problem)



"""Repeat the following paper:
    Loshchilov, I., 2014, July.
    A computationally efficient limited memory CMA-ES for large scale optimization.
    In Proceedings of Annual Conference on Genetic and Evolutionary Computation (pp. 397-404). ACM.
    https://dl.acm.org/doi/abs/10.1145/2576768.2598294

    Original C++/Matlab source code: https://sites.google.com/site/lmcmaeses/

    Given that our code could perfectly generate the *nearly same* results as the original code,
    we argue that the repeatability of `LMCMAES` can be well-documented (*at least partly*).
"""
import pickle

import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

from pypop7.benchmarks.base_functions import ellipsoid
from pypop7.benchmarks.rotated_functions import generate_rotation_matrix
from pypop7.benchmarks.rotated_functions import ellipsoid as rotated_ellipsoid
from pypop7.optimizers.es.lmcmaes import LMCMAES as Solver

sns.set_theme(style='darkgrid')


def plot(function, problem_dim):
    plt.figure()
    result = pickle.load(open(function + '_' + str(problem_dim) + '.pickle', 'rb'))
    plt.xscale('log')
    plt.yscale('log')
    plt.ylim([1e-10, 1e10])
    plt.yticks([1e-10, 1e-8, 1e-6, 1e-4, 1e-2, 1e0, 1e2, 1e4, 1e6, 1e8, 1e10])
    plt.xlim([1e2, 2e8])
    plt.xticks([1e2, 1e3, 1e4, 1e5, 1e6, 1e7, 1e8])
    plt.plot(result['fitness'][:, 0], result['fitness'][:, 1], color='blue')
    plt.xlabel("Function Evaluations")
    plt.ylabel("Objective Function")
    plt.title(function.capitalize())
    plt.show()


if __name__ == '__main__':
    for f in [ellipsoid, rotated_ellipsoid]:
        for dim in [2*128]:
            problem = {'fitness_function': f,
                       'ndim_problem': dim,
                       'lower_boundary': -5.0*np.ones((dim,)),
                       'upper_boundary': 5.0*np.ones((dim,))}
            options = {'fitness_threshold': 1e-10,
                       'seed_rng': 1,  # undefined in the original paper
                       'sigma': 5.0,
                       'verbose_frequency': 2e4,
                       'record_fitness': True,
                       'record_fitness_frequency': 1,
                       'is_restart': False}
            if f in [rotated_ellipsoid]:
                generate_rotation_matrix(f, dim, 0)
                function_name = 'rotated_' + f.__name__
            else:
                function_name = f.__name__
            solver = Solver(problem, options)
            results = solver.optimize()
            print(results['fitness'])
            pickle.dump(results, open(function_name + '_' + str(dim) + '.pickle', 'wb'))
            plot(function_name, dim)



import numpy as np

from pypop7.optimizers.ds.ds import DS


class PriorityQueue(object):
    def __init__(self):
        self.queue = []

    def __len__(self):
        return len(self.queue)

    def __str__(self):
        return ' '.join([str(i) for i in self.queue])

    # for checking if the queue is empty
    def isEmpty(self):
        return len(self.queue) == 0

    # for inserting an element in the queue
    def put(self, data):
        self.queue.append(data)

    # for popping an element based on Priority
    def delete(self):
        try:
            max = 0
            for i in range(len(self.queue)):
                if self.queue[i] > self.queue[max]:
                    max = i
            item = self.queue[max]
            del self.queue[max]
            return item
        except IndexError:
            print()
            exit()


def basis(i, n):
    return np.identity(n)[:, i]


def peek(stack):
    if stack:
        return stack.queue[-1]
    else:
        return None


class Interval:
    def __init__(self, c, y, depths):
        self.c = c
        self.y = y
        self.depths = depths


def reparameterize_to_unit_hypercube(fun, a, b):
    delta = np.subtract(b, a)
    return lambda x: fun(np.multiply(x, delta) + a)


def rev_unit_hypercube_parameterization(x, a, b):
    return np.multiply(x, b - a) + a


def vertex_dist(interval):
    x = 0.5 * 3.0 ** (-interval.depths)
    return np.linalg.norm(x, ord=2)


def min_depth(interval):
    return np.min(interval.depths)


def divide(g, interval):
    c, d, n = np.array(interval.c), min_depth(interval), len(interval.c)
    dirs = np.asarray(np.where(interval.depths == d))[0]
    cs = [(c + np.multiply(3.0 ** (-d - 1), basis(i, n)), c - np.multiply((3.0 ** (-d - 1)), basis(i, n))) for i in
          dirs]
    vs = [(g(C[0]), g(C[1])) for C in cs]
    minvals = [np.min(V) for V in vs]
    intervals = []
    depths = np.copy(interval.depths)
    for j in np.argsort(minvals):
        depths[dirs[j]] += 1
        C, V = cs[j], vs[j]
        intervals.append(Interval(C[0], V[0], np.copy(depths)))
        intervals.append(Interval(C[1], V[1], np.copy(depths)))
    intervals.append(Interval(c, interval.y, np.copy(depths)))
    return intervals


def slope(y, y1, x, x1):
    if x == x1:
        x1 += 1e-10
    return (y1 - y) / (x1 - x)


def get_opt_intervals(intervals, eps):
    stack = []
    for pq in intervals:
        if len(intervals[pq]) != 0:
            interval = peek(intervals[pq])[0]
            x, y = vertex_dist(interval), interval.y
            while len(stack) > 1:
                interval1, interval2 = stack[-1], stack[-2]
                x1, y1 = vertex_dist(interval1), interval1.y
                x2, y2 = vertex_dist(interval2), interval2.y
                l = slope(y, y2, x, x2)
                if y1 <= l * (x1 - x) + y + eps:
                    break
                stack.pop()
            if len(stack) != 0 and y > stack[-1].y + eps:
                continue
            stack.append(interval)
    return stack


class Direct(DS):
    """Divided Rectangles(DIRECT)
        Reference
        ------------
        Kochenderfer, M.J. and Wheeler, T.A., 2019.
        Algorithms for optimization.
        MIT Press.
        https://algorithmsbook.com/optimization/files/chapter-7.pdf
        (See Algorithm 7.8 (Page 117) for details.)
        The original code can be viewed in
        https://github.com/ch4ki/DividedRectangles
        D. R. Jones, C. D. Perttunen, B. E. Stuckman
        A dividing rectangles algorithm for stochastic simulation optimization.
        JOURNAL OF OPTIMIZATION THEORY AND APPLICATION: Vol. 79, No. 1, OCTOBER 1993
        https://link.springer.com/article/10.1007/BF00941892
    """
    def __init__(self, problem, options):
        DS.__init__(self, problem, options)
        self.eps = options.get('eps')
        self.x = None

    def initialize(self):
        g = reparameterize_to_unit_hypercube(self.fitness_function, self.lower_boundary, self.upper_boundary)
        intervals = {}
        # x = self._initialize_x()
        # c = (x - self.lower_boundary) / (self.upper_boundary - self.lower_boundary)
        c = 0.5 * np.ones(self.ndim_problem)
        filling = np.empty(self.ndim_problem)
        filling.fill(0)
        interval = Interval(c, g(c), filling)
        self.add_interval(intervals, interval)
        return g, intervals

    def add_interval(self, intervals, interval):
        d = vertex_dist(interval)
        if d in intervals:
            pass
        else:
            intervals[d] = PriorityQueue()
        intervals[d].put((interval, interval.y))
        self.n_function_evaluations += 1
        if (not self._is_maximization) and (interval.y < self.best_so_far_y):
            self.best_so_far_x, self.best_so_far_y = \
                rev_unit_hypercube_parameterization(interval.c, self.lower_boundary, self.upper_boundary), interval.y
        if self._is_maximization and (-interval.y > self.best_so_far_y):
            self.best_so_far_x, self.best_so_far_y = \
                rev_unit_hypercube_parameterization(interval.c, self.lower_boundary, self.upper_boundary), -interval.y

    def iterate(self, g, intervals):
        s = get_opt_intervals(intervals, self.eps)
        to_add, y = [], []
        for interval in s:
            to_add.append(divide(g, interval))
            del intervals[vertex_dist(interval)]
        for add in to_add:
            for interval in add:
                self.add_interval(intervals, interval)
                y.append(interval.y)
        return intervals, y

    def optimize(self, fitness_function=None):
        fitness = DS.optimize(self, fitness_function)
        g, intervals = self.initialize()
        fitness.append(self.best_so_far_y)
        while True:
            if self._check_terminations():
                break
            intervals, y = self.iterate(g, intervals)
            if self.record_fitness:
                fitness.extend(y)
            self._n_generations += 1
            self._print_verbose_info(y)
        results = self._collect_results(fitness)
        return results



import numpy as np

from pypop7.optimizers.core.optimizer import Optimizer


class EDA(Optimizer):
    """Estimation of Distribution Algorithms (EDA).

    This is the **base** (abstract) class for all EDA classes. Please use any of its concrete subclasses to
    optimize the black-box problem at hand.

    .. note:: Its three methods (`initialize`, `iterate`, `optimize`) should be implemented by all its
       subclasses.

       AKA probabilistic model-building genetic algorithms (PMBGA), iterated density estimation
       evolutionary algorithms (IDEA).

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`),
                * 'record_fitness'           - flag to record fitness list to output results (`bool`, default: `False`),
                * 'record_fitness_frequency' - function evaluations frequency of recording (`int`, default: `1000`),

                  * if `record_fitness` is set to `False`, it will be ignored,
                  * if `record_fitness` is set to `True` and it is set to 1, all fitness generated during optimization
                    will be saved into output results.

                * 'verbose'                  - flag to print verbose info during optimization (`bool`, default: `True`),
                * 'verbose_frequency'        - frequency of printing verbose info (`int`, default: `10`);
              and with two particular settings (`keys`):
                * 'n_individuals' - number of offspring, offspring population size (`int`, default: `200`),
                * 'n_parents'     - number of parents, parental population size (`int`, default:
                  `int(self.n_individuals / 2)`).

    Attributes
    ----------
    n_individuals : `int`
                    number of offspring, offspring population size.
    n_parents     : `int`
                    number of parents, parental population size.

    Methods
    -------

    References
    ----------
    https://www.dagstuhl.de/en/program/calendar/semhp/?semnr=22182

    Larrañaga, P. and Lozano, J.A. eds., 2001.
    Estimation of distribution algorithms: A new tool for evolutionary computation.
    Springer Science & Business Media.
    https://link.springer.com/book/10.1007/978-1-4615-1539-5

    Mühlenbein, H. and Mahnig, T., 2001.
    Evolutionary algorithms: From recombination to search distributions.
    In Theoretical Aspects of Evolutionary Computing (pp. 135-173). Springer, Berlin, Heidelberg.
    https://link.springer.com/chapter/10.1007/978-3-662-04448-3_7

    Berny, A., 2000, September.
    Selection and reinforcement learning for combinatorial optimization.
    In International Conference on Parallel Problem Solving from Nature (pp. 601-610).
    Springer, Berlin, Heidelberg.
    https://link.springer.com/chapter/10.1007/3-540-45356-3_59

    Bosman, P.A. and Thierens, D., 2000, September.
    Expanding from discrete to continuous estimation of distribution algorithms: The IDEA.
    In International Conference on Parallel Problem Solving from Nature (pp. 767-776).
    Springer, Berlin, Heidelberg.
    https://link.springer.com/chapter/10.1007/3-540-45356-3_75

    Mühlenbein, H., 1997.
    The equation for response to selection and its use for prediction.
    Evolutionary Computation, 5(3), pp.303-346.
    https://tinyurl.com/yt78c786
    """
    def __init__(self, problem, options):
        Optimizer.__init__(self, problem, options)
        if self.n_individuals is None:  # number of offspring, offspring population size
            self.n_individuals = 200
        if self.n_parents is None:  # number of parents, parental population size
            self.n_parents = int(self.n_individuals / 2)
        self.n_restart = 0
        self._n_generations = 0

    def initialize(self):
        raise NotImplementedError

    def iterate(self):
        raise NotImplementedError

    def _print_verbose_info(self, y):
        if self.verbose and (not self._n_generations % self.verbose_frequency):
            info = '  * Generation {:d}: best_so_far_y {:7.5e}, min(y) {:7.5e} & Evaluations {:d}'
            print(info.format(self._n_generations, self.best_so_far_y,
                              np.min(y), self.n_function_evaluations))

    def _collect_results(self, fitness):
        results = Optimizer._collect_results(self, fitness)
        results['_n_generations'] = self._n_generations
        results['n_restart'] = self.n_restart
        return results



import numpy as np

from pypop7.optimizers.es.es import ES


class MVAES(ES):
    """Main Vector Adaptation Evolution Strategies (MVAES, MVA-ES).

    Reference
    ---------
    Poland, J. and Zell, A., 2001, July.
    Main vector adaptation: A CMA variant with linear time and space complexity.
    In Proceedings of Annual Conference on Genetic and Evolutionary Computation (pp. 1050-1055).
    https://dl.acm.org/doi/abs/10.5555/2955239.2955428
    """
    def __init__(self, problem, options):
        options['n_parents'] = options.get('n_parents', 1)
        ES.__init__(self, problem, options)
        self.w_v = options.get('w_v', 3)
        self.c_s = options.get('c_s', 4 / (self.ndim_problem + 4))
        self.c_m = options.get('c_m', 4 / (self.ndim_problem + 4))
        self.c_v = options.get('c_v', 2 / np.power(self.ndim_problem + np.sqrt(2), 2))
        # undefined in the original paper (so we use the default value from CMA-ES)
        self.d_s = 1 / self.c_s + 1
        self._e_chi = np.sqrt(self.ndim_problem - 1 / 2)
        self._c_s_1 = 1 - self.c_s
        self._c_s_2 = np.sqrt(self.c_s * (2 - self.c_s))
        self._c_m_1 = 1 - self.c_m
        self._c_m_2 = np.sqrt(self.c_m * (2 - self.c_m))

    def initialize(self, is_restart=False):
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        p_s = np.zeros((self.ndim_problem,))  # evolution path for global step-size adaptation
        p_m = np.zeros((self.ndim_problem,))  # evolution path for main vector adaptation
        v = np.zeros((self.ndim_problem,))  # main (mutation) vector
        z = np.empty((self.n_individuals, self.ndim_problem))
        z_1 = np.empty((self.n_individuals,))
        x = np.empty((self.n_individuals, self.ndim_problem))  # offspring population
        y = np.empty((self.n_individuals,))  # fitness (no evaluation)
        return mean, p_s, p_m, v, z, z_1, x, y

    def iterate(self, mean=None, v=None, z=None, z_1=None, x=None, y=None, args=None):
        for k in range(self.n_individuals):
            if self._check_terminations():
                return z, z_1, x, y
            z[k] = self.rng_optimization.standard_normal((self.ndim_problem,))  # Line 1
            z_1[k] = self.rng_optimization.standard_normal()  # Line 2
            x[k] = mean + self.sigma * (z[k] + z_1[k] * self.w_v * v)  # Line 3
            y[k] = self._evaluate_fitness(x[k], args)
        return z, z_1, x, y

    def _update_distribution(self, p_s=None, p_m=None, v=None, z=None, z_1=None, x=None, y=None):
        order = np.argsort(y)[:self.n_parents]
        mean = np.mean(x[order], axis=0)
        p_s = self._c_s_1 * p_s + self._c_s_2 * np.mean(z[order], axis=0)  # Line 4
        self.sigma *= np.exp((np.linalg.norm(p_s) - self._e_chi) / (self.d_s * self._e_chi))  # Line 5
        p_m_w = np.zeros((self.ndim_problem,))
        for k in range(self.n_parents):
            p_m_w += (z[order[k]] + z_1[order[k]] * self.w_v * v)
        p_m = self._c_m_1 * p_m + self._c_m_2 * p_m_w / self.n_parents  # Line 6
        v = (1 - self.c_v) * np.sign(np.dot(v, p_m)) * v + self.c_v * p_m  # Line 7
        return mean, p_s, p_m, v

    def restart_initialize(self, mean=None, p_s=None, p_m=None, v=None, z=None, z_1=None, x=None, y=None):
        if ES.restart_initialize(self):
            mean, p_s, p_m, v, z, z_1, x, y = self.initialize(True)
        return mean, p_s, p_m, v, z, z_1, x, y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        mean, p_s, p_m, v, z, z_1, x, y = self.initialize()
        while True:
            z, z_1, x, y = self.iterate(mean, v, z, z_1, x, y, args)  # sample and evaluate offspring population
            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            mean, p_s, p_m, v = self._update_distribution(p_s, p_m, v, z, z_1, x, y)
            self._n_generations += 1
            self._print_verbose_info(y)
            if self.is_restart:
                mean, p_s, p_m, v, z, z_1, x, y = self.restart_initialize(mean, p_s, p_m, v, z, z_1, x, y)
        results = self._collect_results(fitness, mean)
        results['v'] = v
        return results



import numpy as np

from pypop7.optimizers.es.es import ES
from pypop7.optimizers.es.opoc import OPOC


class OPOC2009(OPOC):
    """(1+1)-Cholesky-CMA-ES (OPOC).

    Reference
    ---------
    Suttorp, T., Hansen, N. and Igel, C., 2009.
    Efficient covariance matrix update for variable metric evolution strategies.
    Machine Learning, 75(2), pp.167-197.
    https://link.springer.com/article/10.1007/s10994-009-5102-1
    (See Algorithm 2 for details.)
    """
    def __init__(self, problem, options):
        OPOC.__init__(self, problem, options)

    def initialize(self, args=None, is_restart=False):
        mean, y, a, best_so_far_y, p_s = OPOC.initialize(self, args, is_restart)
        p_c = np.zeros((self.ndim_problem,))  # evolution path
        a_i = np.diag(np.ones((self.ndim_problem,)))  # inverse of Cholesky factors
        return mean, y, a, a_i, best_so_far_y, p_s, p_c

    def iterate(self, args=None, mean=None, a=None, a_i=None, best_so_far_y=None, p_s=None, p_c=None):
        # sample and evaluate (only one) offspring
        z = self.rng_optimization.standard_normal((self.ndim_problem,))
        x = mean + self.sigma * np.dot(a, z)
        y = self._evaluate_fitness(x, args)
        if y <= best_so_far_y:
            l_s = 1
        else:
            l_s = 0
        p_s = (1 - self.c_p) * p_s + self.c_p * l_s
        self.sigma *= np.exp(self.eta_sigma * (p_s - self.p_ts) / (1 - self.p_ts))
        if y <= best_so_far_y:
            mean, best_so_far_y = x, y
            if p_s < self.p_t:
                p_c = (1 - self.c_c) * p_c + np.sqrt(self.c_c * (2 - self.c_c)) * np.dot(a, z)
                alpha = 1 - self.c_cov
            else:
                p_c *= 1 - self.c_c
                alpha = 1 - self.c_cov + self.c_cov * self.c_c * (2 - self.c_c)
            beta = self.c_cov
            w = np.dot(a_i, p_c)
            w_norm = np.power(np.linalg.norm(w), 2)
            s_w_norm = np.sqrt(1 + beta / alpha * w_norm)
            a = np.sqrt(alpha) * a + np.sqrt(alpha) / w_norm * (s_w_norm - 1) * np.dot(
                p_c[:, np.newaxis], w[np.newaxis, :])
            a_i = 1 / np.sqrt(alpha) * a_i - 1 / (np.sqrt(alpha) * w_norm) * (1 - 1 / s_w_norm) * np.dot(
                w[:, np.newaxis], np.dot(w[np.newaxis, :], a_i))
        return mean, y, a, a_i, best_so_far_y, p_s, p_c

    def restart_initialize(self, args=None, mean=None, y=None, a=None, a_i=None,
                           best_so_far_y=None, p_s=None, p_c=None, fitness=None):
        self._fitness_list.append(self.best_so_far_y)
        is_restart_1, is_restart_2 = self.sigma < self.sigma_threshold, False
        if len(self._fitness_list) >= self.stagnation:
            is_restart_2 = (self._fitness_list[-self.stagnation] - self._fitness_list[-1]) < self.fitness_diff
        is_restart = bool(is_restart_1) or bool(is_restart_2)
        if is_restart:
            self.n_restart += 1
            self.sigma = np.copy(self._sigma_bak)
            mean, y, a, a_i, best_so_far_y, p_s, p_c = self.initialize(args, is_restart)
            fitness.append(y)
            self._fitness_list = [best_so_far_y]
        return mean, y, a, a_i, best_so_far_y, p_s, p_c

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        mean, y, a, a_i, best_so_far_y, p_s, p_c = self.initialize(args)
        fitness.append(y)
        while True:
            mean, y, a, a_i, best_so_far_y, p_s, p_c = self.iterate(
                args, mean, a, a_i, best_so_far_y, p_s, p_c)
            if self.record_fitness:
                fitness.append(y)
            if self._check_terminations():
                break
            self._n_generations += 1
            self._print_verbose_info(y)
            if self.is_restart:
                mean, y, a, a_i, best_so_far_y, p_s, p_c = self.restart_initialize(
                    args, mean, y, a, a_i, best_so_far_y, p_s, p_c, fitness)
        results = self._collect_results(fitness, mean)
        return results



import time
import numpy as np

from optimizers.nes.nes import NES


class R1NES(NES):
    """Rank-One Natural Evolution Strategy (R1NES).

    Reference
    ---------
    Sun, Y., Schaul, T., Gomez, F. and Schmidhuber, J., 2013, July.
    A linear time natural evolution strategy for non-separable functions.
    In Proceedings of Annual Conference Companion on Genetic and Evolutionary Computation (pp. 61-62).
    https://dl.acm.org/doi/abs/10.1145/2464576.2464608
    https://arxiv.org/abs/1106.1998

    http://schaul.site44.com/code/r1nes.m    (see the official Matlab version)
    """
    def __init__(self, problem, options):
        NES.__init__(self, problem, options)
        if self.eta_mu is None:  # learning rate of mean of Gaussian search distribution (η_μ)
            self.eta_mu = 1
        self.eta_sigma_ = options.get('eta_sigma_', 0.1)
        if options.get('n_individuals') is None:
            self.n_individuals = int(np.floor(8 * np.log(self.ndim_problem)))
        self.upsilon = self._fitness_shaping() + (1 / self.n_individuals)

    def initialize(self):
        s = np.empty((self.n_individuals, self.ndim_problem))  # samples (population / candidate solutions)
        mu = self._initialize_mu()  # mean of Gaussian search distribution
        sigma_ = 0  # [np.exp(sigma_)] == global step-size (sigma)
        # predominant direction: u
        r = 1  # length (== np.exp(c) == ||u||)
        v = self.rng_optimization.standard_normal((self.ndim_problem,))
        v /= np.linalg.norm(v)  # normalized
        u = r * v
        y = np.empty((self.n_individuals,))  # fitness (no evaluation)
        return s, mu, sigma_, r, v, u, y

    def iterate(self, s=None, mu=None, sigma_=None, u=None, y=None, args=None):
        for k in range(self.n_individuals):  # sample population (candidate solutions)
            if self._check_terminations():
                return s, y
            # draw sample (individual)
            s[k] = self.rng_optimization.standard_normal((self.ndim_problem,))
            s[k] += u * self.rng_optimization.standard_normal()
            # evaluate fitness
            y[k] = self._evaluate_fitness(mu + np.exp(sigma_) * s[k], args)
        return s, y

    def _compute_gradients(self, s=None, r=None, v=None, y=None):
        upsilon = np.empty((self.n_individuals,))
        upsilon[np.argsort(y)] = np.copy(self.upsilon)
        g_mu = np.dot(upsilon, s)
        s_dot_s, s_dot_v = np.sum(np.power(s, 2), 1), np.dot(s, v)
        g_sigma_ = (s_dot_s - self.ndim_problem) - (np.power(s_dot_v, 2) - 1)
        g_sigma_ = np.dot(g_sigma_, upsilon) / (2 * (self.ndim_problem - 1))
        g_u = (np.power(r, 2) - self.ndim_problem + 2) * np.power(s_dot_v, 2) - (np.power(r, 2) + 1) * s_dot_s
        g_u /= (2 * r * (self.ndim_problem - 1))
        g_u = np.dot(g_u, upsilon) * v + np.dot(s_dot_v / r * upsilon, s)
        g_c = np.dot(g_u, v) / r
        g_v = g_u / r - g_c * v
        return g_mu, g_sigma_, g_u, g_c, g_v

    def _update_distribution(self, mu=None, sigma_=None, r=None, v=None, u=None,
                             g_mu=None, g_sigma_=None, g_u=None, g_c=None, g_v=None):
        mu += self.eta_mu * np.exp(sigma_) * g_mu
        sigma_ += self.eta_sigma_ * g_sigma_
        # learning rate for r, v, u
        eta = np.minimum(0.1, 2 * np.sqrt(np.power(r, 2) / np.dot(g_u, g_u)))
        if g_c < 0:  # multiplicative update
            r *= np.exp(eta * g_c)  # same as [c += eta_c * g_c] since [r == np.exp(c)]
            v = u / r + eta * g_v
            v /= np.linalg.norm(v)
            u = r * v
        else:  # additive update
            u += eta * g_u
            r = np.linalg.norm(u)  # same as [c = np.log(||u||)] since [r == np.exp(c)]
            v = u / r
        return mu, sigma_, r, v, u

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        self.start_time = time.time()
        fitness = []  # store all fitness generated during evolution
        if fitness_function is not None:
            self.fitness_function = fitness_function
        s, mu, sigma_, r, v, u, y = self.initialize()
        while True:
            s, y = self.iterate(s, mu, sigma_, u, y, args)  # sample and evaluate offspring population
            if self.record_fitness:
                fitness.extend(y.tolist())
            if self._check_terminations():
                break
            g_mu, g_sigma_, g_u, g_c, g_v = self._compute_gradients(s, r, v, y)
            mu, sigma_, r, v, u = self._update_distribution(mu, sigma_, r, v, u, g_mu, g_sigma_, g_u, g_c, g_v)
            self.n_generations += 1
            self._print_verbose_info(y)
        if self.record_fitness:
            self._compress_fitness(fitness[:self.n_function_evaluations])
        results = self._collect_results()
        results['mu'] = mu
        results['sigma'] = np.exp(sigma_)
        results['u'] = u
        return results



import numpy as np

from pypop7.optimizers.rs.rs import RS
from optimizers.sa.csa import CSA


class ESA(CSA):
    """Enhanced Simulated Annealing (ESA).

    Note that this implementation is slightly different from the original paper:
        1. simplify Space Partitioning (Step 2):
            Each dimension can be chosen only once for one movement cycle,
        2. modify Temperature Adjustment (Step 6):
            Temperature should be set properly via hyper-parameter optimization,
        3. delete Four Nonexclusive Stopping Tests (Step 8):
            We use the default stopping conditions suggested by the pypop library.

    Reference
    ---------
    Siarry, P., Berthiau, G., Durdin, F. and Haussy, J., 1997.
    Enhanced simulated annealing for globally minimizing functions of many-continuous variables.
    ACM Transactions on Mathematical Software, 23(2), pp.209-228.
    https://dl.acm.org/doi/abs/10.1145/264029.264043
    """
    def __init__(self, problem, options):
        CSA.__init__(self, problem, options)
        # for Step 5: Test for End of Temperature Stage
        self.N1 = options.get('N1', 12)  # for accepted moves
        self.N2 = options.get('N2', 20)  # for attempted moves
        # for Step 7: Step Vector Adjustment
        self.RATMAX = options.get('RATMAX', 0.2)  # condition to extend step vector
        self.EXTSTP = options.get('EXTSTP', 2)  # extending factor
        self.RATMIN = options.get('RATMIN', 0.05)  # condition to shrink step vector
        self.SHRSTP = options.get('SHRSTP', 0.5)  # shrinking factor
        # parameters at current temperature stage
        self.MVOKST = 0  # number of accepted moves
        self.MOKST = np.zeros((self.ndim_problem,))  # numbers of accepted moves for each dimension
        self.NMVST = 0  # number of attempted moves
        self.MTOTST = np.zeros((self.ndim_problem,))  # numbers of attempted moves for each dimension

    def iterate(self, args=None):  # Step 2, 3, 4
        # Step 2: Space Partitioning
        #  Here we only consider one dimension for each movement and make sure that all
        #  dimensions are optimized in one movement cycle, though their order is randomly generated.
        p = self.rng_optimization.permutation(self.ndim_problem)  # len(p) == n
        fitness = []
        for k in p:
            if self._check_terminations():
                break
            # Step 3: Execution of One Movement
            #   Here we execute one movement for each dimension (a very simple setting).
            x = np.copy(self.parent_x)
            x[k] += self.rng_optimization.uniform(-1, 1) * self.v[k]
            y = self._evaluate_fitness(x, args)
            if self.record_fitness:
                fitness.append(y)
            self._print_verbose_info()
            # Step 4: Acceptance or Rejection of the Movement
            diff = self.parent_y - y
            self.MTOTST[k] += 1
            self.NMVST += 1
            if (diff >= 0) or (self.rng_optimization.random() < np.exp(diff / self.T)):
                self.parent_x, self.parent_y = x, y
                self.MOKST[k] += 1
                self.MVOKST += 1
        return fitness

    def adjust_step_vector(self):  # Step 7: Step Vector Adjustment
        for k in range(self.ndim_problem):
            rok = self.MOKST[k] / self.MTOTST[k]
            if rok > self.RATMAX:
                self.v[k] *= self.EXTSTP
            elif rok < self.RATMIN:
                self.v[k] *= self.SHRSTP

    def reset_parameters(self):  # Step 9: Initialization of a New Temperature Stage
        self.MVOKST = 0
        self.MOKST = np.zeros((self.ndim_problem,))
        self.NMVST = 0
        self.MTOTST = np.zeros((self.ndim_problem,))

    def optimize(self, fitness_function=None, args=None):
        super(RS, self).optimize(fitness_function)
        # Step 1: Initializations
        fitness = [self.initialize(args)]  # store all fitness generated during search
        while not self._check_terminations():
            # Step 5: Test for End of Temperature Stage
            while (self.MVOKST < self.N1 * self.ndim_problem) and (self.NMVST < self.N2 * self.ndim_problem):
                # Step 2, 3, and 4 are combined into `self.iterate`:
                fitness.extend(self.iterate(args))
                if self._check_terminations():
                    break
            # Step 6: Temperature Adjustment
            #  Here we don't execute the original paper, which is complicated (fitness-dependent).
            #  Instead, we execute a very simple version from [Corana et al., 1987] and expect that
            #  in practice such a hyper-parameter will be properly set by systematic search.
            self.T *= self.r_T
            # Step 7: Step Vector Adjustment
            self.adjust_step_vector()
            # Step 8: Four Nonexclusive Stopping Tests
            #  Here we don't execute them, since their settings are problem-dependent.
            #  In practice, it appears to be difficult to provide problem-independent executions.
            # Step 9: Initialization of a New Temperature Stage
            self.reset_parameters()
        return self._collect_results(fitness)
